;;;; -*- lisp -*-

(in-package :it.bese.yaclml)

(defun make-tal-enviroment ()
  (make-hash-table :test 'equalp))

(defun lookup-tal-variable (name enviroment)
  (gethash name enviroment))

(defun (setf lookup-tal-variable) (value name enviroment)
  (setf (gethash name enviroment) value))

(defun read-tal-expression-from-string (expression enviroment-symbol)
  (let ((*readtable* (copy-readtable nil)))
    (declare (special *readtable*))
    (set-macro-character #\$ (lambda (stream char)
                               (declare (ignore char))
                               `(lookup-tal-variable ,(string (read stream)) ,enviroment-symbol))
                         nil *readtable*)
    (read-from-string expression)))

(defun normalize-lhtml (lhtml)
  "Convert the LHTML code into a more consistent form."
  (etypecase lhtml
    (keyword (list (list lhtml)))
    (string lhtml)
    (list (let ((tag-name (car lhtml))
                (body (cdr lhtml)))
            (cons (ensure-list tag-name) body)))))

(defun handle-tal-when (tag output-stream-symbol enviroment-symbol)
  (let ((value (getf  (cdr (car tag)) :TAL\:WHEN)))
    (remf (cdr (car tag)) :TAL\:WHEN)
    `(when ,(read-tal-expression-from-string value enviroment-symbol)
       ,(transform-lhtml-tree (list tag) output-stream-symbol enviroment-symbol))))

(defun handle-tal-unless (tag output-stream-symbol enviroment-symbol)
  (let ((value (getf (cdr (car tag)) :TAL\:UNLESS)))
    (remf (cdr (car tag)) :TAL\:UNLESS)
    `(unless ,(read-tal-expression-from-string value enviroment-symbol)
       ,(transform-lhtml-tree (list tag) output-stream-symbol enviroment-symbol))))

(defun transform-lhtml-tree (tree output-stream-symbol enviroment-symbol)
  "Given a tree representing some LHTML code with TAL attributes
  returns the code for generating the HTML.

  Returns a progn with the appropiate WRITE-STRING
  STATEMENTS. Destructivly modifies TREE."
  (with-collector (ops)
    (labels ((walk-lhtml (lhtml)
               (dolist (l (mapcar #'normalize-lhtml lhtml))
                 (if (stringp l)
                     (ops l)
                     (destructuring-bind ((tag-name &rest attributes) &rest body) l
                       (cond-bind value
                         ((getf attributes :TAL\:WHEN)
                          (ops (handle-tal-when l output-stream-symbol enviroment-symbol)))
                         ((getf attributes :TAL\:UNLESS)
                          (ops (handle-tal-unless l output-stream-symbol enviroment-symbol)))
                         (t (ops "<" (string tag-name))
                            (loop
                               for (name value) on attributes by #'cddr
                               do (ops " " (string name) "=\"" value "\""))
                            (if body
                                (progn
                                  (ops ">")
                                  (walk-lhtml body)
                                  (ops "</" (string tag-name) ">"))
                                (ops "/>")))))))))
      (walk-lhtml tree)
      `(progn ,@(mapcar (lambda (form)
                          (if (stringp form)
                              `(write-line ,form ,output-stream-symbol)
                              form))
                        (fold-strings (ops)))))))

(defun compile-tal-string (string)
  (compile nil
           (let ((output-stream (gensym))
                 (enviroment (gensym)))
             `(lambda (,output-stream ,enviroment)
                (declare (ignorable ,enviroment))
                ,(transform-lhtml-tree (parse-html string)
                                       output-stream
                                       enviroment)))))
